/*
 * cryptofs.c -- the Simple Block Utility
 *
 * Copyright (C) 2001 Alessandro Rubini and Jonathan Corbet
 * Copyright (C) 2001 O'Reilly & Associates
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 * $Id: cryptofs.c,v 1.32 2001/07/18 22:28:16 rubini Exp $
 */

#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

#include <linux/config.h>
#include <linux/module.h>

#include <linux/sched.h>
#include <linux/kernel.h> /* printk() */
#include <linux/malloc.h> /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/timer.h>
#include <linux/types.h>  /* size_t */
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/hdreg.h>  /* HDIO_GETGEO */

#include <linux/errno.h>  /* error codes */
int errno;


#include <asm/unistd.h>
_syscall3(int,write,int,fd,const char *,buf,off_t,count);
_syscall3(int,read,int,fd,char *,buf,off_t,count);
_syscall3(off_t,lseek,int,fd,off_t,offset,int,count);
_syscall3(int,open,const char *,file,int,flag,int,mode);
_syscall1(int,close,int,fd);
_syscall0(uid_t,getuid);
_syscall2(int,ftruncate,int,fd,off_t,length);

#include <asm/system.h>   /* cli(), *_flags */

#define MAJOR_NR cryptofs_major /* force definitions on in blk.h */
static int cryptofs_major; /* must be declared before including blk.h */

#define DEVICE_NR(device) MINOR(device)   /* cryptofs has no partition bits */
#define DEVICE_NAME "cryptofs"               /* name for messaging */
#define DEVICE_INTR cryptofs_intrptr         /* pointer to the bottom half */
#define DEVICE_NO_RANDOM                  /* no entropy to contribute */
#define DEVICE_REQUEST cryptofs_request
#define DEVICE_OFF(d) /* do-nothing */

#include <linux/blk.h>

#include "cryptofs.h"        /* local definitions */
#ifdef HAVE_BLKPG_H
#include <linux/blkpg.h>  /* blk_ioctl() */
#endif

/*
 * Do the raw char interface in 2.4.
 */
#ifdef LINUX_24
#  define DO_RAW_INTERFACE
#  include <linux/iobuf.h>
static void cryptofsr_init();
static void cryptofsr_release();
#  define cryptofsR_SECTOR 512  /* insist on this */
#  define cryptofsR_SECTOR_MASK (cryptofsR_SECTOR - 1)
#  define cryptofsR_SECTOR_SHIFT 9
#endif

/*
 * Non-prefixed symbols are static. They are meant to be assigned at
 * load time. Prefixed symbols are not static, so they can be used in
 * debugging. They are hidden anyways by register_symtab() unless
 * cryptofs_DEBUG is defined.
 */
static int major    = cryptofs_MAJOR;
static int devs     = cryptofs_DEVS;
static int rahead   = cryptofs_RAHEAD;
static int size     = cryptofs_SIZE;
static int blksize  = cryptofs_BLKSIZE;
static int hardsect = cryptofs_HARDSECT;

MODULE_PARM(major, "i");
MODULE_PARM(devs, "i");
MODULE_PARM(rahead, "i");
MODULE_PARM(size, "i");
MODULE_PARM(blksize, "i");
MODULE_PARM(hardsect, "i");
MODULE_AUTHOR("Alessandro Rubini");

static mm_segment_t old_limit;
#define limitoff   old_limit = get_fs(); set_fs(KERNEL_DS)
#define limiton    set_fs(old_limit)


int cryptofs_devs, cryptofs_rahead, cryptofs_size;
int cryptofs_blksize, cryptofs_hardsect;

/* The following items are obtained through kmalloc() in cryptofs_init() */

cryptofs_Dev *cryptofs_devices = NULL;
int *cryptofs_blksizes = NULL;
int *cryptofs_sizes = NULL;
int *cryptofs_hardsects = NULL;

/*
 * We can do without a request queue, but only in 2.4
 */
#if defined(LINUX_24) && !defined(cryptofs_MULTIQUEUE)
static int noqueue = 0;         /* Use request queue by default */
MODULE_PARM(noqueue, "i");
#endif

#ifdef DO_RAW_INTERFACE
int cryptofsr_major = cryptofsR_MAJOR;
MODULE_PARM(cryptofsr_major, "i");
#endif


int cryptofs_revalidate(kdev_t i_rdev);


/*
 * Open and close
 */

int cryptofs_open (struct inode *inode, struct file *filp)
{
    cryptofs_Dev *dev; /* device information */
    int num = MINOR(inode->i_rdev);

    if (num >= cryptofs_devs) return -ENODEV;
        
    dev = cryptofs_devices + num;
    
    /* kill the timer associated to the device: it might be active */
    del_timer(&dev->timer);


    spin_lock(&dev->lock);
    /* revalidate on first open and fail if no data is there */
    if (!dev->usage) {
        check_disk_change(inode->i_rdev);
        if (!dev->data)
        {
            spin_unlock (&dev->lock);
            return -ENOMEM;
        }
    }
printk("<1>I in c_open: num = %i\n",num);    
    if( dev->filename )
    {
     limitoff;
     printk("<1>I in c_open: num = %i, fs=%s, key=%s\n",num,dev->filename,dev->key);        
     dev->fd = open(dev->filename,O_RDWR|O_CREAT,0600);

     if(dev->fd<0){ limiton; spin_unlock(&dev->lock); return -errno;}
     printk("<1>I in c_open: final\n");    
     limiton;    
    }
    
    dev->usage++;
    spin_unlock(&dev->lock);
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int cryptofs_release (struct inode *inode, struct file *filp)
{
    cryptofs_Dev *dev = cryptofs_devices + MINOR(inode->i_rdev);

    spin_lock(&dev->lock);
    dev->usage--;
    /*
     * If the device is closed for the last time, start a timer
     * to release RAM in half a minute. The function and argument
     * for the timer have been setup in cryptofs_init()
     */
    if (!dev->usage) {
        dev->timer.expires = jiffies + 30 * HZ;
        add_timer(&dev->timer);
        /* but flush it right now */
        fsync_dev(inode->i_rdev);
        invalidate_buffers(inode->i_rdev);
    }
    
    if( dev->fd>=0 )
    {
     close(dev->fd);
     dev->fd = -1;
    }

    
    MOD_DEC_USE_COUNT;
    spin_unlock(&dev->lock);
    return 0;
}


/*
 * The timer function. As argument it receives the device
 */
void cryptofs_expires(unsigned long data)
{
    cryptofs_Dev *dev = (cryptofs_Dev *)data;

    spin_lock(&dev->lock);
    if (dev->usage || !dev->data) {
        spin_unlock(&dev->lock);
        printk(KERN_WARNING "cryptofs: timer mismatch for device %i\n",
               dev - cryptofs_devices);
        return;
    }
    PDEBUG("freeing device %i\n",dev - cryptofs_devices);
    vfree(dev->data);
    dev->data=0;
    spin_unlock(&dev->lock);
    return;
}    

/*
 * The ioctl() implementation
 */

int cryptofs_ioctl (struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
    int err;
    long size;
    struct hd_geometry geo;
    struct attribute* atr;
    cryptofs_Dev* dev;

    PDEBUG("ioctl 0x%x 0x%lx\n", cmd, arg);
    switch(cmd) {
     case 1000:
        atr = (struct attribute*) arg;
	printk("<1>ioctl: %s %s %i\n",atr->filename,atr->key,atr->size);
	dev = cryptofs_devices + MINOR(inode->i_rdev);
	if(dev->filename) kfree(dev->filename), kfree(dev->key);
	dev->filename = kmalloc(strlen(atr->filename)+1,GFP_KERNEL);
	if(!dev->filename) return -ENOMEM;
	dev->key = kmalloc(strlen(atr->key)+1,GFP_KERNEL);
	if(!dev->key) return -ENOMEM;	
	strcpy(dev->filename,atr->filename);
	strcpy(dev->key,atr->key);
	printk("<1>all ok! %s %s %i\n",dev->filename,dev->key,atr->size);
	return 0;
     
      case BLKGETSIZE:
        /* Return the device size, expressed in sectors */
        if (!arg) return -EINVAL; /* NULL pointer: not valid */
        err = ! access_ok (VERIFY_WRITE, arg, sizeof(long));
        if (err) return -EFAULT;
        size = blksize*cryptofs_sizes[MINOR(inode->i_rdev)]
		/ cryptofs_hardsects[MINOR(inode->i_rdev)];
	if (copy_to_user((long *) arg, &size, sizeof (long)))
	    return -EFAULT;
        return 0;

      case BLKRRPART: /* re-read partition table: can't do it */
        return -ENOTTY;

      case HDIO_GETGEO:
        /*
	 * Get geometry: since we are a virtual device, we have to make
	 * up something plausible.  So we claim 16 sectors, four heads,
	 * and calculate the corresponding number of cylinders.  We set the
	 * start of data at sector four.
         */
        err = ! access_ok(VERIFY_WRITE, arg, sizeof(geo));
        if (err) return -EFAULT;
        size = cryptofs_size * blksize / cryptofs_hardsect;
        geo.cylinders = (size & ~0x3f) >> 6;
	geo.heads = 4;
	geo.sectors = 16;
	geo.start = 4;
	if (copy_to_user((void *) arg, &geo, sizeof(geo)))
	    return -EFAULT;
        return 0;

      default:
        /*
         * For ioctls we don't understand, let the block layer handle them.
         */
        return blk_ioctl(inode->i_rdev, cmd, arg);
    }

    return -ENOTTY; /* unknown command */
}

/*
 * Support for removable devices
 */

int cryptofs_check_change(kdev_t i_rdev)
{
    int minor = MINOR(i_rdev);
    cryptofs_Dev *dev = cryptofs_devices + minor;

    if (minor >= cryptofs_devs) /* paranoid */ 
        return 0;                           
printk("check_change for dev %i\n",minor); 
    PDEBUG("check_change for dev %i\n",minor);
    if (dev->data)
        return 0; /* still valid */
	printk("<1>check_change for dev %i expired\n",minor); 
    return 1; /* expired */
}


/*
 * Note no locks taken out here.  In a worst case scenario, we could drop
 * a chunk of system memory.  But that should never happen, since validation
 * happens at open or mount time, when locks are held.
 */
int cryptofs_revalidate(kdev_t i_rdev)
{
    cryptofs_Dev *dev = cryptofs_devices + MINOR(i_rdev);

printk("<1>revalidate for dev %i\n",MINOR(i_rdev));

    PDEBUG("revalidate for dev %i\n",MINOR(i_rdev));
    if (dev->data)
        return 0;
    dev->data = vmalloc(dev->size);
    if (!dev->data)
        return -ENOMEM;
    return 0;
}

/*
 * The file operations
 */

#ifdef LINUX_24
struct block_device_operations cryptofs_bdops = {
    open:               cryptofs_open,
    release:            cryptofs_release,
    ioctl:              cryptofs_ioctl,
    check_media_change: cryptofs_check_change,
    revalidate:         cryptofs_revalidate,
};
#else

#ifdef LINUX_20
void cryptofs_release_20 (struct inode *inode, struct file *filp)
{
        (void) cryptofs_release (inode, filp);
}
#define cryptofs_release cryptofs_release_20
#endif

struct file_operations cryptofs_bdops = {
    read:       block_read,
    write:      block_write,
    ioctl:      cryptofs_ioctl,
    open:       cryptofs_open,
    release:    cryptofs_release,
    fsync:      block_fsync,
    check_media_change: cryptofs_check_change,
    revalidate: cryptofs_revalidate
};
# endif /* LINUX_24 */




/*
 * Block-driver specific functions
 */

/*
 * Find the device for this request.
 */
static cryptofs_Dev *cryptofs_locate_device(const struct request *req)
{
    int devno;
    cryptofs_Dev *device;

    /* Check if the minor number is in range */
    devno = DEVICE_NR(req->rq_dev);
    if (devno >= cryptofs_devs) {
        static int count = 0;
        if (count++ < 5) /* print the message at most five times */
            printk(KERN_WARNING "cryptofs: request for unknown device\n");
        return NULL;
    }
    device = cryptofs_devices + devno;  /* Pick it out of our device array */
    return device;
}


/*
 * Perform an actual transfer.
 */
static int cryptofs_transfer(cryptofs_Dev *device, const struct request *req)
{
    int size;
    u8 *ptr;
    unsigned int ptr1;
    int i, pos=0, keylen;
    int devno;
    
    ptr1 = req->sector * cryptofs_hardsect;
    ptr = device->data + req->sector * cryptofs_hardsect;
    size = req->current_nr_sectors * cryptofs_hardsect;
    devno = DEVICE_NR(req->rq_dev);
    
    /* Make sure that the transfer fits within the device. */
    if (ptr + size > device->data + cryptofs_blksize*cryptofs_size) {
        static int count = 0;
        if (count++ < 5)
            printk(KERN_WARNING "cryptofs: request past end of device\n");
        return 0;
    }

    if( device->fd<0 ) return 0;
    /* Looks good, do the transfer. */
    switch(req->cmd) {
        case READ:
	    lseek(device->fd,ptr1,0);
	    limitoff;
	     i = read(device->fd,req->buffer,size);
	printk("<1>I in transfer, read: %i %i err: %i\n",ptr1,i,errno);
/*	     keylen = strlen(device->key);
	     for(i=0; i<cryptofs_blksizes[devno]; i++) // 
	     {
	      req->buffer[i] ^= device->key[pos];
	      pos = ++pos == keylen ? 0: pos;
	     } */
	    limiton;
	    if( i<0 ) return 0;
            //memcpy(req->buffer, ptr, size); /* from cryptofs to buffer */
            return 1;
        case WRITE:
	    lseek(device->fd,ptr1,0);
	    limitoff;
/*	    keylen = strlen(device->key);
	     for(i=0; i<cryptofs_blksizes[devno]; i++) // 
	     {
	      req->buffer[i] ^= device->key[pos];
	      pos = ++pos == keylen ? 0: pos;
	     } */
	     i = write(device->fd,req->buffer,size);
	printk("<1>I in transfer, write: %i %i err: %i\n",ptr1,i,errno);
	    limiton;
            //memcpy(ptr, req->buffer, size); /* from buffer to cryptofs */
	    if( i<0 ) return 0;
            return 1;
        default:
            /* can't happen */
            return 0;
    }
}


#ifdef LINUX_24

/*
 * Transfer a buffer directly, without going through the request queue.
 */
int cryptofs_make_request(request_queue_t *queue, int rw, struct buffer_head *bh)
{
    u8 *ptr;

    /* Figure out what we are doing */
    cryptofs_Dev *device = cryptofs_devices + MINOR(bh->b_rdev);
    ptr = device->data + bh->b_rsector * cryptofs_hardsect;

    /* Paranoid check, this apparently can really happen */
    if (ptr + bh->b_size > device->data + cryptofs_blksize*cryptofs_size) {
        static int count = 0;
        if (count++ < 5)
            printk(KERN_WARNING "cryptofs: request past end of device\n");
        bh->b_end_io(bh, 0);
        return 0;
    }

    /* This could be a high memory buffer, shift it down */
#if CONFIG_HIGHMEM
    bh = create_bounce(rw, bh);
#endif

    /* Do the transfer */
    switch(rw) {
        case READ:
        case READA:  /* Readahead */
            memcpy(bh->b_data, ptr, bh->b_size); /* from cryptofs to buffer */
            bh->b_end_io(bh, 1);
            break;
        case WRITE:
            refile_buffer(bh);
            memcpy(ptr, bh->b_data, bh->b_size); /* from buffer to cryptofs */
            mark_buffer_uptodate(bh, 1);
            bh->b_end_io(bh, 1);
            break;
        default:
            /* can't happen */
            bh->b_end_io(bh, 0);
            break;
    }

    /* Nonzero return means we're done */
    return 0;
}


void cryptofs_unused_request(request_queue_t *q)
{
    static int count = 0;
    if (count++ < 5)
        printk(KERN_WARNING "cryptofs: unused_request called\n");
}
        
#endif /* LINUX_24 */



#if defined(cryptofs_EMPTY_REQUEST)
/*
 * This empty request function just prints the interesting items
 * of the current request. The sectors affected by the request
 * are printed as <first-sector>-<number-of-sectors>.
 */
#ifdef LINUX_24
void cryptofs_request(request_queue_t *q)
#else                                   
void cryptofs_request()                    
#endif                                  
{
    while(1) {
        INIT_REQUEST;
        printk("<1>request %p: cmd %i sec %li (nr. %li)\n", CURRENT,
               CURRENT->cmd,
               CURRENT->sector,
               CURRENT->current_nr_sectors);
        end_request(1); /* success */
    }
}


#elif defined(cryptofs_MULTIQUEUE)  /* 2.4 only */

/*
 * Clean up this request.
 */
int cryptofs_end_request(struct request *req, int status)
{
    if (end_that_request_first(req, status, DEVICE_NAME))
        return 1;
    end_that_request_last(req);
    return 0;
}



void cryptofs_request(request_queue_t *q)
{
    cryptofs_Dev *device;
    struct request *req;
    int status;

    /* Find our device */
    device = cryptofs_locate_device (blkdev_entry_next_request(&q->queue_head));
    if (device->busy) /* no race here - io_request_lock held */
        return;
    device->busy = 1;

    /* Process requests in the queue */
    while(! list_empty(&q->queue_head)) {

    /* Pull the next request off the list. */
        req = blkdev_entry_next_request(&q->queue_head);
        blkdev_dequeue_request(req);
        spin_unlock_irq (&io_request_lock);
        spin_lock(&device->lock);

    /* Process all of the buffers in this (possibly clustered) request. */
        do {
            status = cryptofs_transfer(device, req);
        } while (end_that_request_first(req, status, DEVICE_NAME));
        spin_unlock(&device->lock);
        spin_lock_irq (&io_request_lock);
        end_that_request_last(req);
    }
    device->busy = 0;
}

/*
 * Tell the block layer where to queue a request.
 */
request_queue_t *cryptofs_find_queue(kdev_t device)
{
    int devno = DEVICE_NR(device);

    if (devno >= cryptofs_devs) {
        static int count = 0;
        if (count++ < 5) /* print the message at most five times */
            printk(KERN_WARNING "cryptofs: request for unknown device\n");
        return NULL;
    }
    return &cryptofs_devices[devno].queue;
}



#else /* not cryptofs_MULTIQUEUE */
#ifdef LINUX_24                                 
void cryptofs_request(request_queue_t *q)
#else                                           
void cryptofs_request()                            
#endif                                          
{
    cryptofs_Dev *device;
    int status;

    while(1) {
        INIT_REQUEST;  /* returns when queue is empty */

        /* Which "device" are we using? */
        device = cryptofs_locate_device (CURRENT);
        if (device == NULL) {
            end_request(0);
            continue;
        }

        /* Perform the transfer and clean up. */
	spin_lock(&device->lock);
        status = cryptofs_transfer(device, CURRENT);
        spin_unlock(&device->lock);
        end_request(status); 
    }
}
#endif /* not cryptofs_EMPTY_REQUEST nor cryptofs_MULTIQUEUE */

/*
 * Finally, the module stuff
 */

int cryptofs_init(void)
{
    int result, i;

    /*
     * Copy the (static) cfg variables to public prefixed ones to allow
     * snoozing with a debugger.
     */

    cryptofs_major    = major;
    cryptofs_devs     = devs;
    cryptofs_rahead   = rahead;
    cryptofs_size     = size;
    cryptofs_blksize  = blksize;
    cryptofs_hardsect = hardsect;

#ifdef LINUX_20
    /* Hardsect can't be changed :( */
    if (hardsect != 512) {
        printk(KERN_ERR "cryptofs: can't change hardsect size\n");
        hardsect = cryptofs_hardsect = 512;
    }
#endif
    /*
     * Register your major, and accept a dynamic number
     */
    result = register_blkdev(cryptofs_major, "cryptofs", &cryptofs_bdops);
    if (result < 0) {
        printk(KERN_WARNING "cryptofs: can't get major %d\n",cryptofs_major);
        return result;
    }
    if (cryptofs_major == 0) cryptofs_major = result; /* dynamic */
    major = cryptofs_major; /* Use `major' later on to save typing */

    /*
     * Assign the other needed values: request, rahead, size, blksize,
     * hardsect. All the minor devices feature the same value.
     * Note that `cryptofs' defines all of them to allow testing non-default
     * values. A real device could well avoid setting values in global
     * arrays if it uses the default values.
     */

    read_ahead[major] = cryptofs_rahead;
    result = -ENOMEM; /* for the possible errors */

    cryptofs_sizes = kmalloc(cryptofs_devs * sizeof(int), GFP_KERNEL);
    if (!cryptofs_sizes)
        goto fail_malloc;
    for (i=0; i < cryptofs_devs; i++) /* all the same size */
        cryptofs_sizes[i] = cryptofs_size;
    blk_size[major]=cryptofs_sizes;

    cryptofs_blksizes = kmalloc(cryptofs_devs * sizeof(int), GFP_KERNEL);
    if (!cryptofs_blksizes)
        goto fail_malloc;
    for (i=0; i < cryptofs_devs; i++) /* all the same blocksize */
        cryptofs_blksizes[i] = cryptofs_blksize;
    blksize_size[major]=cryptofs_blksizes;

    cryptofs_hardsects = kmalloc(cryptofs_devs * sizeof(int), GFP_KERNEL);
    if (!cryptofs_hardsects)
        goto fail_malloc;
    for (i=0; i < cryptofs_devs; i++) /* all the same hardsect */
        cryptofs_hardsects[i] = cryptofs_hardsect;
    hardsect_size[major]=cryptofs_hardsects;

    /* FIXME: max_readahead and max_sectors */ 
    
    /* 
     * allocate the devices -- we can't have them static, as the number
     * can be specified at load time
     */

    cryptofs_devices = kmalloc(cryptofs_devs * sizeof (cryptofs_Dev), GFP_KERNEL);
    if (!cryptofs_devices)
        goto fail_malloc;
    memset(cryptofs_devices, 0, cryptofs_devs * sizeof (cryptofs_Dev));
    for (i=0; i < cryptofs_devs; i++) {
        cryptofs_devices[i].fd = -1;
        /* data and usage remain zeroed */
        cryptofs_devices[i].size = 1024 * cryptofs_size;
        init_timer(&(cryptofs_devices[i].timer));
        cryptofs_devices[i].timer.data = (unsigned long)(cryptofs_devices+i);
        cryptofs_devices[i].timer.function = cryptofs_expires;
        spin_lock_init(&cryptofs_devices[i].lock);
    }

    /*
     * Get the queue set up, and register our (nonexistent) partitions.
     */  
#ifdef cryptofs_MULTIQUEUE
    for (i = 0; i < cryptofs_devs; i++) {
        blk_init_queue(&cryptofs_devices[i].queue, cryptofs_request);
        blk_queue_headactive(&cryptofs_devices[i].queue, 0);
    }
    blk_dev[major].queue = cryptofs_find_queue;
#else
#  ifdef LINUX_24
    if (noqueue)
        blk_queue_make_request(BLK_DEFAULT_QUEUE(major), cryptofs_make_request);
    else
#  endif /* LINUX_24 */
        blk_init_queue(BLK_DEFAULT_QUEUE(major), cryptofs_request);
#endif
    /* A no-op in 2.4.0, but all drivers seem to do it anyway */
    for (i = 0; i < cryptofs_devs; i++)
            register_disk(NULL, MKDEV(major, i), 1, &cryptofs_bdops,
                            cryptofs_size << 1);

#ifndef cryptofs_DEBUG
    EXPORT_NO_SYMBOLS; /* otherwise, leave global symbols visible */
#endif

    printk ("<1>cryptofs: init complete, %d devs, size %d blks %d hs %d\n",
                    cryptofs_devs, cryptofs_size, cryptofs_blksize, cryptofs_hardsect);
#ifdef cryptofs_MULTIQUEUE
    printk ("<1>cryptofs: Using multiqueue request\n");
#elif defined(LINUX_24)
    if (noqueue)
            printk (KERN_INFO "cryptofs: using direct make_request\n");
#endif

#ifdef DO_RAW_INTERFACE
    cryptofsr_init();
#endif
    return 0; /* succeed */

  fail_malloc:
    read_ahead[major] = 0;
    if (cryptofs_sizes) kfree(cryptofs_sizes);
    blk_size[major] = NULL;
    if (cryptofs_blksizes) kfree(cryptofs_blksizes);
    blksize_size[major] = NULL;
    if (cryptofs_hardsects) kfree(cryptofs_hardsects);
    hardsect_size[major] = NULL;
    if (cryptofs_devices) kfree(cryptofs_devices);

    unregister_blkdev(major, "cryptofs");
    return result;
}

void cryptofs_cleanup(void)
{
    int i;

/*
 * Before anything else, get rid of the timer functions.  Set the "usage"
 * flag on each device as well, under lock, so that if the timer fires up
 * just before we delete it, it will either complete or abort.  Otherwise
 * we have nasty race conditions to worry about.
 */
    for (i = 0; i < cryptofs_devs; i++) {
        cryptofs_Dev *dev = cryptofs_devices + i;
        del_timer(&dev->timer);
        spin_lock(&dev->lock);
        dev->usage++;
        spin_unlock(&dev->lock);
    }

#ifdef DO_RAW_INTERFACE
    cryptofsr_release();
#endif
    
    /* flush it all and reset all the data structures */


    for (i=0; i<cryptofs_devs; i++)
        fsync_dev(MKDEV(cryptofs_major, i)); /* flush the devices */
    unregister_blkdev(major, "cryptofs");
/*
 * Fix up the request queue(s)
 */
#ifdef cryptofs_MULTIQUEUE
    for (i = 0; i < cryptofs_devs; i++)
            blk_cleanup_queue(&cryptofs_devices[i].queue);
    blk_dev[major].queue = NULL;
#else
    blk_cleanup_queue(BLK_DEFAULT_QUEUE(major));
#endif   

    /* Clean up the global arrays */
    read_ahead[major] = 0;
    kfree(blk_size[major]);
    blk_size[major] = NULL;
    kfree(blksize_size[major]);
    blksize_size[major] = NULL;
    kfree(hardsect_size[major]);
    hardsect_size[major] = NULL;
    /* FIXME: max_readahead and max_sectors */ 


    /* finally, the usual cleanup */
    for (i=0; i < cryptofs_devs; i++) {
        if (cryptofs_devices[i].data)
            vfree(cryptofs_devices[i].data);
        if( cryptofs_devices[i].filename ) kfree(cryptofs_devices[i].filename);
	if( cryptofs_devices[i].key ) kfree(cryptofs_devices[i].key);
    }
    kfree(cryptofs_devices);
}



/*
 * Below here is the "raw device" implementation, available only
 * in 2.4.
 */
#ifdef DO_RAW_INTERFACE

/*
 * Transfer an iovec
 */
static int cryptofsr_rw_iovec(cryptofs_Dev *dev, struct kiobuf *iobuf, int rw,
                int sector, int nsectors)
{
    struct request fakereq;
    struct page *page;
    int offset = iobuf->offset, ndone = 0, pageno, result;

    /* Perform I/O on each sector */
    fakereq.sector = sector;
    fakereq.current_nr_sectors = 1;
    fakereq.cmd = rw;
    
    for (pageno = 0; pageno < iobuf->nr_pages; pageno++) {
        page = iobuf->maplist[pageno];
        while (ndone < nsectors) {
            /* Fake up a request structure for the operation */
            fakereq.buffer = (void *) (kmap(page) + offset);
            result = cryptofs_transfer(dev, &fakereq);
	    kunmap(page);
            if (result == 0)
                return ndone;
            /* Move on to the next one */
            ndone++;
            fakereq.sector++;
            offset += cryptofsR_SECTOR;
            if (offset >= PAGE_SIZE) {
                offset = 0;
                break;
            }
        }
    }
    return ndone;
}


/*
 * Handle actual transfers of data.
 */
static int cryptofsr_transfer (cryptofs_Dev *dev, char *buf, size_t count,
                loff_t *offset, int rw)
{
    struct kiobuf *iobuf;       
    int result;
    
    /* Only block alignment and size allowed */
    if ((*offset & cryptofsR_SECTOR_MASK) || (count & cryptofsR_SECTOR_MASK))
        return -EINVAL;
    if ((unsigned long) buf & cryptofsR_SECTOR_MASK)
        return -EINVAL;

    /* Allocate an I/O vector */
    result = alloc_kiovec(1, &iobuf);
    if (result)
        return result;

    /* Map the user I/O buffer and do the I/O. */
    result = map_user_kiobuf(rw, iobuf, (unsigned long) buf, count);
    if (result) {
        free_kiovec(1, &iobuf);
        return result;
    }
    spin_lock(&dev->lock);
    result = cryptofsr_rw_iovec(dev, iobuf, rw, *offset >> cryptofsR_SECTOR_SHIFT,
                    count >> cryptofsR_SECTOR_SHIFT);
    spin_unlock(&dev->lock);

    /* Clean up and return. */
    unmap_kiobuf(iobuf);
    free_kiovec(1, &iobuf);
    if (result > 0)
        *offset += result << cryptofsR_SECTOR_SHIFT;
    return result << cryptofsR_SECTOR_SHIFT;
}


/*
 * Read and write syscalls.
 */
ssize_t cryptofsr_read(struct file *filp, char *buf, size_t size, loff_t *off)
{
    cryptofs_Dev *dev = cryptofs_devices + MINOR(filp->f_dentry->d_inode->i_rdev);
    return cryptofsr_transfer(dev, buf, size, off, READ);
}

ssize_t cryptofsr_write(struct file *filp, const char *buf, size_t size,
                loff_t *off)
{
    cryptofs_Dev *dev = cryptofs_devices + MINOR(filp->f_dentry->d_inode->i_rdev);
    return cryptofsr_transfer(dev, (char *) buf, size, off, WRITE);
}


static int cryptofsr_registered = 0;
static struct file_operations cryptofsr_fops = {
   read:        cryptofsr_read,
   write:       cryptofsr_write,
   open:        cryptofs_open,
   release:     cryptofs_release,
   ioctl:	cryptofs_ioctl,
};


static void cryptofsr_init()
{
    int result;
        
    /* Simplify the math */
    if (cryptofs_hardsect != cryptofsR_SECTOR) {
        printk(KERN_NOTICE "cryptofsr requires hardsect = %d\n", cryptofsR_SECTOR);
        return;
    }
    SET_MODULE_OWNER(&cryptofsr_fops);
    result = register_chrdev(cryptofsr_major, "cryptofsr", &cryptofsr_fops);
    if (result >= 0)
        cryptofsr_registered = 1;
    if (cryptofsr_major == 0)
        cryptofsr_major = result;
}


static void cryptofsr_release()
{
    if (cryptofsr_registered)
        unregister_chrdev(cryptofsr_major, "cryptofsr");
}




#endif /* DO_RAW_INTERFACE */

module_init(cryptofs_init);
module_exit(cryptofs_cleanup);
